import logging
import re
from pathlib import Path
from typing import List

from .command import Commands, Move
from .base_filename_cleaner import BaseFilenameCleaner


class DuplicateStringRemover(BaseFilenameCleaner):

    def fix_commands(self, min_length: int, recurse: bool) -> Commands:
        fix_commands: Commands = []
        if recurse:
            subdirs = sorted([name for name in self._base_directory.iterdir() if name.is_dir()])
            for subdir in subdirs:
                cleaner = DuplicateStringRemover(self._base_directory.joinpath(subdir.name))
                fix_commands += cleaner.fix_commands(min_length, recurse)

        files = self.get_music_files()
        to_remove = self.longest_common_substring_in_filenames(files)
        to_remove = self._exclude_common_use_cases(to_remove)
        if min_length and len(to_remove) < min_length:
            return fix_commands
        logging.info('----    DIR: %s    ----    REMOVE: "%s"', self._base_directory, to_remove)
        for file in files:
            fix_commands.append(Move(
                self._base_directory.joinpath(file),
                self._base_directory.joinpath(file.name.replace(to_remove, '')),
                [to_remove]
            ))
        return fix_commands

    @staticmethod
    def longest_common_substring_in_filenames(data: List[Path]) -> str:
        return DuplicateStringRemover.longest_common_substring([f.name for f in data])

    @staticmethod
    def longest_common_substring(data: List[str]) -> str:
        substr = ''
        if len(data) > 1 and data[0]:
            for i in range(len(data[0])):
                for j in range(len(data[0]) - i + 1):
                    if j > len(substr) and all(data[0][i:i + j] in x for x in data):
                        substr = data[0][i:i + j]
        return substr

    @staticmethod
    def _exclude_common_use_cases(to_remove: str) -> str:
        if to_remove[:3] == ' - ':
            to_remove = to_remove[3:]
        if re.match(r'.*\.\w+$', to_remove):
            to_remove = re.sub(r'\.\w+?$', '', to_remove)
        if to_remove.endswith('-'):
            to_remove = to_remove[:-1]
        elif to_remove.startswith('-'):
            to_remove = to_remove[1:]
        return to_remove



# pylint: disable=too-many-arguments
__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import urllib.request
from collections import Counter
from datetime import datetime
from json import loads
from sys import maxsize
from typing import List, Dict

from media_tools.util.buying.track import Track


class LastFM:

    API_BASE = 'http://ws.audioscrobbler.com/2.0/'

    def __init__(self, user: str, api_key: str) -> None:
        self.user = user
        self.api_key = api_key

    def base_url(self, method: str, limit: int) -> str:
        return f'{self.API_BASE}?method={method}&user={self.user}&api_key={self.api_key}&' \
            f'limit={limit}&format=json'

    def get_top_tracks(
            self, period: str = 'overall', limit: int = 50, min_plays: int = 1, page: int = 1
    ) -> List[Track]:
        url = f'{self.base_url("user.gettoptracks", limit * page)}&period={period}'
        tracks = self.get_response(url, 'toptracks')['track']
        top_tracks: List[Track] = [None] * len(tracks)
        for track in tracks:
            top_tracks[int(track['@attr']['rank']) - 1] = Track.from_dict(track)
        return [
            track for track in top_tracks if track.playcount >= min_plays
        ][limit * (page - 1):limit * page]

    def get_tracks_by_period(
            self, from_date: datetime, to_date: datetime = datetime.now(), limit: int = 50,
            min_plays: int = 1, page: int = 1
    ) -> List[Track]:
        url = f'{self.base_url("user.getrecenttracks", 200)}&' \
            f'from={from_date.timestamp()}&to={to_date.timestamp()}'
        tracks = Counter(self.get_tracks_paged(url, 'recenttracks'))
        for track, playcount in tracks.items():
            track.playcount = playcount
        return [
            track for track, num_plays in tracks.most_common(limit * page) if num_plays >= min_plays
        ][limit * (page - 1):limit * page]

    def get_tracks_paged(self, url: str, method: str) -> List[Track]:
        tracks = []
        page, total_pages = 1, maxsize
        while page <= total_pages:
            response = self.get_response(f'{url}&page={page}', method)
            total_pages = int(response['@attr']['totalPages'])
            page = int(response['@attr']['page']) + 1
            tracks.append(response['track'])
        return [Track.from_dict(t) for chunk in tracks for t in chunk]

    @staticmethod
    def get_response(url: str, method: str) -> Dict:
        with urllib.request.urlopen(url) as response:
            return loads(response.read())[method]



import re
from pathlib import Path
from typing import List, Tuple, Dict

from .command import Command, Relink, Nothing, Commands
from .base_filename_cleaner import BaseFilenameCleaner
from .junk_remover import JunkRemover
from .numbering_fixer import NumberingFixer
from .util import find_potential_files


class SymlinkFixer(BaseFilenameCleaner):

    def fix_commands(self, undo_info: Dict[str, str]) -> Commands:
        symlinks = self.find_broken_symlinks_with_original_targets()
        return [self._fix_symlink(symlink, target, undo_info) for symlink, target in symlinks]

    def _fix_symlink(self, symlink: Path, old_target: Path, undo_info: Dict[str, str]) -> Command:
        to_try = self.patterns_to_try(old_target, undo_info)
        for file in to_try:
            filepath = Path(file)
            if filepath.is_file():
                return Relink(symlink, filepath)
        return Nothing()

    def find_broken_symlinks_with_original_targets(self) -> List[Tuple[Path, Path]]:
        return [(symlink, self.symlink_target(symlink)) for symlink in self.find_broken_symlinks()]

    def find_broken_symlinks(self) -> List[Path]:
        def is_broken_symlink(path: Path) -> bool:
            return path.is_symlink() and not path.is_file()

        return find_potential_files(self._base_directory, is_broken_symlink)

    @staticmethod
    def symlink_target(symlink: Path) -> Path:
        return symlink.resolve()

    def patterns_to_try(self, moved_file: Path, undo_info: Dict[str, str]) -> List[Path]:
        to_try = []

        for source, dest in undo_info.items():
            if str(moved_file) in source:
                to_try.append(Path(dest))

        junk = JunkRemover(self._base_directory)
        to_try.append(junk.fix_command_for_file(moved_file)[1])

        for file in [moved_file] + to_try:
            for pattern in NumberingFixer.PATTERNS_TO_FIX:
                match = re.match("(.*)/" + pattern, str(file))
                if match:
                    fixed = f"{match.group(1)}/{match.group(2)} - {match.group(3)}"
                    to_try.append(Path(fixed))

        return to_try



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from pathlib import Path
from typing import Tuple, Optional

from .authorization_error import AuthorizationError
from .constants import (
    MP3_KBIT_RATE, DEFAULT_CROSSFADE_MS, MIXCLOUD_MAX_FILESIZE, ACCESS_TOKEN_FILE,
    ACCESS_TOKEN_SEARCH_PATH
)
from .mix import Mix
from .multi_mix import MultiMix


def bytes_per_second(mp3_kbit_rate: int = MP3_KBIT_RATE) -> int:
    return mp3_kbit_rate // 8 * 1024 * 2


def create_mix(  # pylint: disable=too-many-arguments
        basedir: Path, patterns: Tuple[str, ...], access_token: str,
        strict: bool = False, crossfade_ms: int = DEFAULT_CROSSFADE_MS, title: str = None
) -> "Mix":
    files, length = Mix.scan(basedir, patterns)
    if bytes_per_second() * length < MIXCLOUD_MAX_FILESIZE:
        return Mix(basedir, patterns, access_token, strict, crossfade_ms, title)
    return MultiMix(
        basedir, files, access_token, length, strict, crossfade_ms, title
    )


def get_access_token(access_token_path: Optional[Path] = None) -> str:
    error_message = f"""
Authorization token does not exist - please follow the instructions at
https://www.mixcloud.com/developers/#authorization to generate an auth
token and store it under ./{ACCESS_TOKEN_FILE}, ~/{ACCESS_TOKEN_FILE}
or ~/.config/{ACCESS_TOKEN_FILE}.
"""
    if not access_token_path:
        for basedir in ACCESS_TOKEN_SEARCH_PATH:
            if (basedir / ACCESS_TOKEN_FILE).exists():
                access_token_path = basedir / ACCESS_TOKEN_FILE
                break
    if access_token_path is None:
        raise AuthorizationError(error_message)
    try:
        with access_token_path.open('r') as file:
            return file.read().strip()
    except OSError as error:
        raise AuthorizationError(error_message) from error



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import json
import logging
import pickle
import sys
from argparse import ArgumentParser
from bz2 import open as bzopen
from calendar import timegm
from datetime import date, datetime
from glob import glob
from gzip import open as gzopen
from lzma import open as xzopen
from os import environ
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Union

from pylast import LastFMNetwork, User, PyLastError

from media_tools.util.logging import setup_logging

MAX_RETRY_WAIT = 1024
OPEN_FUNCTIONS: Dict[str, Callable] = {
    'bz2': bzopen,
    'gz': gzopen,
    'xz': xzopen,
}
DUMP_PARAMS = {
    'json': dict(dump=json.dump, mode='wt', kwargs=dict(default=str)),
    'pickle': dict(dump=pickle.dump, mode='wb', kwargs={})
}


def parse():
    parser = ArgumentParser(
        description="Creates a pickle or JSON file from a last.fm user's data"
    )
    parser.add_argument(
        '-l', '--limit', type=int, default=None,
        help='Max. number of entries to request'
    )
    parser.add_argument(
        '-y', '--year', type=int, default=None,
        help='Year the data is requested for'
    )
    parser.add_argument(
        '-c', '--combine', action='store_true',
        help='Combine data sets present in current folder'
    )
    parser.add_argument(
        '-o', '--output', type=str, default=None,
        help='output file name'
    )
    args = parser.parse_args()
    return args


def from_datetime(convert_datetime: datetime) -> int:
    return timegm(convert_datetime.utctimetuple())


def year_range(year: Optional[int]) -> Dict[str, int]:
    start = None if year is None else from_datetime(datetime(year, 1, 1, 0, 0, 0))
    end = None if year is None else from_datetime(datetime(year, 12, 31, 23, 59, 59, 999999))
    return dict(time_from=start, time_to=end)


def start_year(user: User) -> int:
    return date.fromtimestamp(user.get_unixtime_registered()).year


def get_and_wait(
        values: Optional[List], function: Callable, wait_time: int, **kwargs
) -> Union[List[Any], Any]:
    if values is None:
        values = function(**kwargs)
        logging.info('waiting %ss', str(wait_time))
        sleep(wait_time)
    return values


def get_data(  # pylint: disable=too-many-arguments
        user: User, year: Optional[int], limit: Optional[int], wait_time: int = 0,
        tracks: Optional[List[str]] = None, loved: Optional[List[str]] = None,
        top_tracks: Optional[List[str]] = None
) -> Dict:
    logging.info("get_data(user=%s, year=%s, limit=%s, wait=%s)", user.name, year, limit, wait_time)
    year_dates = year_range(year)
    try:
        play_count = user.get_playcount()
        tracks = get_and_wait(tracks, user.get_recent_tracks, wait_time, limit=limit, **year_dates)
        loved = get_and_wait(loved, user.get_loved_tracks, wait_time, limit=limit)
        top_tracks = get_and_wait(top_tracks, user.get_top_tracks, wait_time, limit=limit)
        top_artists = get_and_wait(None, user.get_top_artists, wait_time, limit=limit)
        top_albums = get_and_wait(None, user.get_top_albums, wait_time, limit=limit)
        top_tags = user.get_top_tags(limit=limit)
        return dict(
            time=datetime.now(),
            play_count=play_count,
            tracks=tracks,
            loved=loved,
            top_tracks=top_tracks,
            top_artists=top_artists,
            top_albums=top_albums,
            top_tags=top_tags,
        )
    # pylast has a feature to deal with rate limiting, but it doesn't seem to work reliably
    except PyLastError as error:
        logging.info("%s: %s", type(error).__name__, str(error))
        if wait_time <= MAX_RETRY_WAIT:
            next_wait_time = 1 if wait_time == 0 else 2 * wait_time
            logging.warning("%s, retrying with %ss wait time", type(error).__name__, next_wait_time)
            return get_data(
                user, year, limit, next_wait_time,
                tracks=tracks, loved=loved, top_tracks=top_tracks
            )
        logging.error('Rate limiting still active after %ss - giving up', MAX_RETRY_WAIT)
        raise


def combine_data() -> None:
    separate_user_data = {}
    user_data = {}
    for filename in glob('backup_lastfm_20??.pickle'):
        with open(filename, 'rb') as file:
            separate_user_data[filename] = pickle.load(file)
        user_data = separate_user_data[filename]
    user_data['tracks'] = sum([data['tracks'] for data in separate_user_data.values()], [])
    with open('backup_lastfm_all.pickle', 'wb') as backup_file:
        pickle.dump(user_data, backup_file)


def get_output_filename(year: Optional[int], filename: Optional[str]) -> str:
    if filename is None:
        year_string = "all" if year is None else str(year)
        filename = f'backup_lastfm_{year_string}.pickle.bz2'
    return filename


def get_and_dump(
        user: User, year: Optional[int], limit: Optional[int], filename: Optional[str] = None
):
    user_data = get_data(user, year, limit)
    backup_filename = get_output_filename(year, filename)
    dump(user_data, backup_filename)


def dump(user_data, backup_filename):
    filename_parts = backup_filename.split('.')
    extension = filename_parts[-1].lower()
    dump_format = extension if extension in DUMP_PARAMS else filename_parts[-2].lower()
    dump_params = DUMP_PARAMS.get(dump_format, DUMP_PARAMS['pickle'])
    dump_func = dump_params['dump']
    mode = dump_params['mode']
    kwargs = dump_params['kwargs']
    with OPEN_FUNCTIONS.get(extension, open)(backup_filename, mode) as backup_file:
        dump_func(user_data, backup_file, **kwargs)


def main() -> None:
    """
    Usage:
    $ export PYLAST_USERNAME=ENTER_YOURS_HERE
    $ export PYLAST_API_KEY=ENTER_YOURS_HERE
    $ export PYLAST_API_SECRET=ENTER_YOURS_HERE
    $ python backup_lastfm_data [options]
    """
    args = parse()
    setup_logging(args)

    if args.combine:
        combine_data()
        sys.exit(0)

    network = LastFMNetwork(
        api_key=environ['PYLAST_API_KEY'], api_secret=environ['PYLAST_API_SECRET']
    )
    user = network.get_user(environ['PYLAST_USERNAME'])

    if args.year is None:
        get_and_dump(user, None, args.limit, args.output)
    else:
        years = [args.year] if args.year else range(start_year(user), datetime.now().year)
        for year in years:
            get_and_dump(user, year, args.limit, args.output)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['media_tools',
 'media_tools.util',
 'media_tools.util.buying',
 'media_tools.util.mixcloud']

package_data = \
{'': ['*']}

install_requires = \
['audioread>=2.1.9',
 'coloredlogs>=15.0.1',
 'lxml>=4.8.0',
 'mutagen>=1.45.1',
 'pydub>=0.24.1',
 'pylast>=5.0.0',
 'python-magic>=0.4.25',
 'requests>=2.27.1',
 'tinytag>=1.8.1']

entry_points = \
{'console_scripts': ['backup_lastfm_data = media_tools.backup_lastfm_data:main',
                     'buy_most_played = media_tools.buy_most_played:main',
                     'clean_filenames = media_tools.clean_filenames:main',
                     'copy_from_playlist = media_tools.copy_from_playlist:main',
                     'mixcloud_upload = media_tools.mixcloud_upload:main',
                     'print_length = media_tools.print_length:main']}

setup_kwargs = {
    'name': 'media-tools',
    'version': '0.2.8',
    'description': 'Creating and managing playlists, and managing the filenames and directory structure for large numbers of music files.',
    'long_description': '-   [Installation](#installation)\n-   [Usage](#usage)\n    -   [`mixcloud_upload`](#mixcloud_upload)\n        -   [Authorization](#authorization)\n        -   [Configuration options](#configuration-options)\n    -   [`clean_filenames`](#clean_filenames)\n        -   [Checking results](#checking-results)\n    -   [Audacious playlist tools](#audacious-playlist-tools)\n        -   [`copy_from_playlist`](#copy_from_playlist)\n    -   [`buy_most_played`](#buy_most_played)\n    -   [`backup_lastfm_data`](#backup_lastfm_data)\n    -   [`print_length`](#print_length)\n-   [Development](#development)\n    -   [Test suite](#test-suite)\n    -   [Generating a Table of Contents to\n        `README.md`](#generating-a-table-of-contents-to-readme.md)\n\nToolbox for helping creating and managing playlists, and managing the filenames and directory\nstructure for large numbers of music files.\n\nThis is mostly tailored to my own usage patterns, although it may be useful to others.\n\nHome page: https://gitlab.com/lilacashes/music-library-tools\n\nPyPI page: https://pypi.org/project/media-tools\n\nInstallation\n============\n\n``` {.bash}\n$ pip install media-tools\n```\n\nUsage\n=====\n\n`mixcloud_upload`\n-----------------\n\nGenerate a mix on [Mixcloud](mixcloud.com) from the contents of a local directory.\n\nAssumes that the given directory contains a number of audio files as well as a picture file (JPEG or\nPNG format), one file called `description.txt` which contains the description which is printed along\nwith the mix on Mixcloud, and a file called `tags.txt` which contains a newline-separated list of\ntags for the mix. The mix is created as an audio file called `mix.mp3` in the given directory and\nthen uploaded to Mixcloud. The title of the mix on Mixcloud will be the name of the directory given,\nminus any leading numbers.\n\nExample: You have a directory structured like this:\n\n    10 - My awesome mix\n    |-- 01 - Awesome track.mp3\n    |-- 02 - Also awesome track.ogg\n    |-- 03 - Another pretty cool track.m4a\n    |-- 04 - This track rocks.flac\n    |-- cover.png\n    |-- description.txt\n    |-- tags.txt\n\n``` {.shell}\n$ mixcloud_upload -d "10 - My awesome mix"\n```\n\nwill upload this mix to Mixcloud under the name "My awesome mix" with track 01 to 04 in that order.\n\n### Authorization\n\nYou will need to generate an application on Mixcloud for your account and get an access token, as\ndescribed in the [Mixcloud API Documentation](https://www.mixcloud.com/developers/#authorization).\nStore this access token as `.mixcloud_access_token` in either the current directory, your user\'s\n`$HOME` directory or under `$HOME/.config/media-tools`.\n\n### Configuration options\n\n-   `-d DIRECTORY`, `--directory DIRECTORY`: Specify the folder containing the mix, as seen above\n-   `-e EXTENSIONS [EXTENSIONS ...]`, `--extensions EXTENSIONS [EXTENSIONS ...]`: List of file\n    extensions to consider as audio files for the mix. Per default, files ending in `.mp3`, `.MP3`,\n    `.flac`, `.m4a` and `.ogg` are treated as audio and added to the mix. Change this list of some\n    of your audio files have extensions that are not part of this list, or if you want to exclude\n    some of these extensions.\n-   `-q`, `--quiet`: Only errors are printed. Usually you want to omit `-q`, since the entire\n    generating and uploading process takes several minutes, so you want to be sure what is going on.\n-   `-s`, `--strict`: Do not upload the mix if any of these conditions are met:\n    -   There is no picture, no `descriptions.txt` or no `tags.txt` in the upload folder\n    -   Any of the audio tracks in the folder is missing the ID3 tag for title or artist. These tags\n        are needed to generate the tracklist for the mix.\n-   `-c CROSSFADE_MS`, `--crossfade-ms CROSSFADE_MS`: Number of milliseconds to use for crossfading\n    between two tracks. Currently, there can only be one global value which is used between all\n    tracks.\n-   `-r MAX_RETRY`, `--max-retry MAX_RETRY`: Maximum number of retries for failing uploads. Uploads\n    can fail for a number of reasons, such as rate limiting on the Mixcloud side, or because of a\n    bad internet connection. The default number of retries is set to 100, set this to a lower number\n    if you want to give up earlier.\n-   `-a`, `--auth-token-file`: explicitly specify the file containing the Mixcloud authorization\n    token.\n-   `-t`, `--auth-token-string`: explicitly specify the Mixcloud authorization token as a string.\n\n`clean_filenames`\n-----------------\n\nIn general: run a script with the `-v` option first to see what it would change. If satisfied, then\nre-run it with the `-f` option to effect those changes.\n\nRemoving useless duplicate strings from filenames:\n\n``` {.bash}\n$ clean_filenames -v clean-filenames --recurse .\n$ clean_filenames -f clean-filenames --recurse .\n```\n\nChanging filenames from different numbering schemes to the scheme `01 - filename.ext`:\n\n``` {.bash}\n$ clean_filenames -v clean-numbering .\n$ clean_filenames -f clean-numbering .\n```\n\nRemoving stray junk, such as underscores, stray dashes, and stray `[]` and `()` from filenames:\n\n``` {.bash}\n$ clean_filenames -v clean-junk .\n$ clean_filenames -f clean-junk .\n```\n\nUndoing renamings done with this script, limited to a specified directory and its subfolders, or a\nsingle file name:\n\n``` {.bash}\n$ clean_filenames -v undo .\n$ clean_filenames -f undo .\n# OR\n$ clean_filenames -f undo ./subdir/file_name.mp3\n```\n\nFixing symlinks to files which have been renamed by any of the previous commands:\n\n``` {.bash}\n$ clean_filenames -v fix-symlinks .\n$ clean_filenames -f fix-symlinks .\n```\n\n### Checking results\n\nFinding mp3 files which do not conform to the numbering scheme in general:\n\n``` {.bash}\n$ find . -name \\*.mp3 | grep -vE \'[[:digit:]]+ - .+\\.mp3\'\n```\n\nFinding mp3 files which have a number in their filename but do not conform to the numbering scheme,\nexcluding some more common use cases:\n\n``` {.bash}\n$ find . -name \\*.mp3 | \\\n    grep -E \'[[:digit:]]+[^/]+\\.mp3\' | \\\n    grep -vE \'[[:digit:]]+ - .+\\.mp3\' | \\\n    grep -vE \'[[:digit:]]{2}\\.mp3\'\n```\n\nAudacious playlist tools\n------------------------\n\nTools for making and repairing playlists containing physical music files from audacious playlists\nand the latest music files. The argument to `--playlist` defaults to the playlist currently playing\nin audacious.\n\n### `copy_from_playlist`\n\nCopying files from the current or specified playlist (as its name in the `playlists` subdir in the\naudacious configuration folder) to a specified target folder, optionally limiting the number of\nfiles to copy to the first NUM, and optionally renaming the files to reflect the position of the\nsong in the playlist:\n\n``` {.bash}\n$ copy_from_playlist [-v] copy \\\n    [--playlist PLAYLIST_ID] \\\n    [--number NUM] \\\n    [--renumber] \\\n    TARGET_DIR\n```\n\nTry to find files in the current playlist which are unavailable because they have been moved, and\nmove them back to the place in the filesystem which is noted on the playlist (does not appear to\nwork currently):\n\n``` {.bash}\n$ copy_from_playlist [-v] restore \\\n    [--playlist PLAYLIST_ID]\n```\n\nCopy the newest files to a specified target:\n\n``` {.bash}\n$ copy_from_playlist [-v] copy-newest \\\n    --max-age NUM_DAYS \\\n    --source SOURCE_DIR \\\n    --target TARGET_DIR\n```\n\n`buy_most_played`\n-----------------\n\nAttempt to buy your most played tracks (in a given period) from several platforms.\n\n...\n\n`backup_lastfm_data`\n--------------------\n\nBack up Last.FM scrobbles.\n\n...\n\n`print_length`\n--------------\n\nGiven a directory of audio files, print the total playing length of that directory and all\ndirectories below.\n\n...\n\nDevelopment\n===========\n\nAfter cloning, it is recommended to set up the git hook that runs the test suite before every\n`git push`:\n\n``` {.shell}\n$ cd .git/hooks\n$ ln -s ../../.git_hooks/pre-push .\n```\n\nTest suite\n----------\n\nRun all in one:\n\n``` {.shell}\n$ .git_hooks/pre-push\n```\n\nRun separately:\n\n``` {.shell}\n$ poetry run mypy .\n$ poetry run flake8 .\n$ poetry run pylint media_tools\n$ poetry run nosetests --with-coverage --cover-package=media_tools.util tests/unit\n$ poetry run nosetests tests/integration\n```\n\nGenerating a Table of Contents to `README.md`\n---------------------------------------------\n\n``` {.shell}\n$ git add README.md\n$ git commit -m \'Before ToC generation\'\n$ pandoc -s --columns 100 --toc --toc-depth=4 README.md -o README_toc.md\n$ mv README_toc.md README.md\n$ git add README.md\n$ git commit -m \'After ToC generation\'\n```\n',
    'author': 'Lene Preuss',
    'author_email': 'lene.preuss@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/lilacashes/music-library-tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
import re
import sys
from pathlib import Path
from time import sleep
from typing import Tuple, List, Dict, Optional

import requests
from pydub import AudioSegment
from tinytag import TinyTag

from media_tools.util.mixcloud.constants import (
    DEFAULT_CROSSFADE_MS, DEFAULT_MAX_RETRY, MP3_KBIT_RATE, MIXCLOUD_API_UPLOAD_URL
)


class Mix:  # pylint: disable=too-many-instance-attributes

    @classmethod
    def scan(cls, basedir: Path, patterns: Tuple[str, ...]) -> Tuple[List[str], int]:
        audio_files = sorted([f for p in patterns for f in basedir.glob(p)])
        length = sum([TinyTag.get(str(audio_file)).duration for audio_file in audio_files])
        return [f.name for f in audio_files], int(length)

    def __init__(  # pylint: disable=too-many-arguments
            self, basedir: Path, patterns: Tuple[str, ...], access_token: str,
            strict: bool = False, crossfade_ms: int = DEFAULT_CROSSFADE_MS, title: str = None
    ) -> None:
        self._basedir = basedir
        self._strict = strict
        self._crossfade_ms = crossfade_ms
        self._title = title
        self._incomplete = False
        self._track_info: List[Dict] = []
        self._access_token = access_token
        audio_files = sorted([f for p in patterns for f in self._basedir.glob(p)])
        self._audio = self._import_audio(audio_files)

    @property
    def valid(self):
        return not self._incomplete

    @property
    def tags(self) -> List[str]:
        if (self._basedir / 'tags.txt').exists():
            with (self._basedir / 'tags.txt').open() as file:
                tags = list(line for line in (line.strip() for line in file) if line)
            if len(tags) <= 5:
                return tags
            self._incomplete = True
            raise ValueError(f'Max. 5 tags allowed, found {len(tags)}: {tags}')
        self._incomplete = True
        if self._strict:
            raise ValueError('No tags found')
        logging.warning('No tags found')
        return ['Testing Mixcloud API']

    @property
    def title(self) -> str:
        if self._title:
            return self._title
        title = re.sub(r'^\d+ - ', '', self._basedir.resolve().name)
        return f"Test - don't bother playing ({title})" if self._incomplete else title

    @property
    def description(self) -> str:
        if (self._basedir / 'description.txt').exists():
            with (self._basedir / 'description.txt').open() as file:
                return file.read().strip()
        if self._strict:
            raise ValueError('No description found')
        logging.warning('No description found')
        self._incomplete = True
        return 'Test test test'

    @property
    def picture(self) -> Optional[Path]:
        """Currently just returns the first JPEG or PNG. Room for improvement!"""
        try:
            return next(self._basedir.glob('*.*p*g'))
        except StopIteration as error:
            if self._strict:
                raise ValueError(f"No picture in {self._basedir}") from error
            logging.warning('No picture found')
            self._incomplete = True
            return None

    def _import_audio(self, audio_files: List[Path]) -> AudioSegment:
        audio = AudioSegment.empty()
        for i, audio_file in enumerate(audio_files):
            self._track_info.append(self._get_track_info(audio_file))
            logging.info("%s - %s", i + 1, audio_file.name)
            track = AudioSegment.from_file(audio_file)
            track = track.normalize()
            audio = audio.append(
                track,
                crossfade=self._crossfade_ms if len(audio) > self._crossfade_ms else len(audio)
            )

        return audio

    def _get_track_info(self, audio_file: Path) -> Dict:
        tags = TinyTag.get(str(audio_file))
        if tags.artist is None or tags.title is None:
            if self._strict:
                raise ValueError(f"Incomplete tags for {audio_file}")
            logging.warning("Incomplete tags for %s", audio_file)
            self._incomplete = True
            return {
                'artist': tags.artist or '???', 'title': tags.title or '???',
                'length': tags.duration
            }
        return {
            'artist': tags.artist, 'title': tags.title, 'length': tags.duration,
            'filename': audio_file.name
        }

    def export(self, name: Path = Path('mix.mp3')) -> None:
        mix_file = self._basedir / name
        audio_format = name.suffix[1:]
        logging.info('Exporting to %s with bitrate %s kbps', mix_file, MP3_KBIT_RATE)
        self._audio.export(
            mix_file, format=audio_format, parameters=["-q:a", "0"], bitrate=f'{MP3_KBIT_RATE}k'
        )

    def upload(self, name: Path = Path('mix.mp3'), max_retry: int = DEFAULT_MAX_RETRY) -> None:
        url = MIXCLOUD_API_UPLOAD_URL + '/?access_token=' + self._access_token
        mix_file = self._basedir / name
        if not mix_file.exists():
            raise FileNotFoundError(mix_file)
        files = {
            'mp3': ('mix.mp3', mix_file.open('rb'), 'audio/mpeg'),
        }
        if self.picture:
            picture_type = self.picture.suffix
            files['picture'] = (
                'picture' + picture_type, self.picture.open('rb'),
                f'image/{"png" if picture_type == ".png" else "jpeg"}'
            )

        data = {
            'name': self.title,
            'description': self.description,
            'percentage_music': 100
        }
        self._add_tags(data)
        self._add_track_info(data)
        size_bytes = mix_file.stat().st_size + self.picture.stat().st_size if self.picture else 0
        logging.info(
            'Uploading %s kBytes (%s minutes) as %s',
            f'{size_bytes // 1024:,d}',
            f'{len(self._audio) // 60000}:{len(self._audio) % 60000 // 1000:02}',
            self.title
        )

        try:
            response = requests.post(url, files=files, data=data)
        except requests.exceptions.ConnectionError as error:
            logging.warning(self.connection_error_message(error, max_retry))
            if max_retry > 0:
                self.upload(name, max_retry=max_retry - 1)
                return
            sys.exit(1)
        else:
            logging.info(self.response_message(response))
            if self.is_rate_limited(response) and max_retry > 0:
                sleep(int(response.json()['error']['retry_after']))
                self.upload(name, max_retry=max_retry - 1)
            else:
                sys.exit(1)

    @staticmethod
    def is_rate_limited(response: requests.Response) -> bool:
        return response.status_code == 403 and \
            response.json().get('error', {}).get('type') == 'RateLimitException'

    @staticmethod
    def response_message(response: requests.Response) -> str:
        if response.status_code == 200:
            return f"{response.status_code}: {response.json()['result']['message']}"
        return f"{response.status_code}: {response.json()['error']}"

    @staticmethod
    def connection_error_message(error: requests.exceptions.ConnectionError, max_retry: int) -> str:
        message = ""
        if error.args:
            if error.args[0].args and len(error.args[0].args) > 1:
                message = f"{error.args[0].args[1].args[1]}: {error.args[0].args[0]} "
            else:
                message = str(error.args)
        if max_retry > 0:
            message += f"Retrying {max_retry} time{'s' if max_retry > 1 else ''}."
        else:
            message += "Giving up."
        return message

    def _add_tags(self, data: Dict) -> None:
        for i, tag in enumerate(self.tags):
            data[f"tags-{i}-tag"] = tag

    def _add_track_info(self, data: Dict) -> None:
        start_time = 0
        for i, track_info in enumerate(self._track_info):
            data[f"sections-{i}-artist"] = track_info['artist']
            data[f"sections-{i}-song"] = track_info['title']
            data[f"sections-{i}-start_time"] = int(start_time)
            start_time += (track_info['length'] - self._crossfade_ms / 1000)
